@@ -1,27 +1,26 @@
 --- Generally useful routines.
 -- See  @{01-introduction.md.Generally_useful_functions|the Guide}.
+--
+-- Dependencies: `pl.compat`
+--
 -- @module pl.utils
-local format,gsub,byte = string.format,string.gsub,string.byte
-local clock = os.clock
+local format = string.format
+local compat = require 'pl.compat'
 local stdout = io.stdout
 local append = table.insert
+local unpack = rawget(_G,'unpack') or rawget(table,'unpack')
 
-local collisions = {}
-
-local utils = {}
-
-utils._VERSION = "1.1.0"
-
-local lua51 = rawget(_G,'setfenv')
-
-utils.lua51 = lua51
-if not lua51 then -- Lua 5.2 compatibility
-    unpack = table.unpack
-    loadstring = load
-end
-
-utils.dir_separator = _G.package.config:sub(1,1)
-local is_windows = utils.dir_separator == '\\'
+local utils = {
+    _VERSION = "1.5.2",
+    lua51 = compat.lua51,
+    setfenv = compat.setfenv,
+    getfenv = compat.getfenv,
+    load = compat.load,
+    execute = compat.execute,
+    dir_separator = compat.dir_separator,
+    is_windows = compat.is_windows,
+    unpack = unpack
+}
 
 --- end this program gracefully.
 -- @param code The exit code or a message to be printed
@@ -59,7 +58,8 @@ local function import_symbol(T,k,v,libname)
     local key = rawget(T,k)
     -- warn about collisions!
     if key and k ~= '_M' and k ~= '_NAME' and k ~= '_PACKAGE' and k ~= '_VERSION' then
-        utils.printf("warning: '%s.%s' overrides existing symbol\n",libname,k)
+        utils.fprintf(io.stderr,"warning: '%s.%s' will not override existing symbol\n",libname,k)
+        return
     end
     rawset(T,k,v)
 end
@@ -123,24 +123,30 @@ local raise
 function utils.readfile(filename,is_bin)
     local mode = is_bin and 'b' or ''
     utils.assert_string(1,filename)
-    local f,err = io.open(filename,'r'..mode)
-    if not f then return utils.raise (err) end
-    local res,err = f:read('*a')
+    local f,open_err = io.open(filename,'r'..mode)
+    if not f then return utils.raise (open_err) end
+    local res,read_err = f:read('*a')
     f:close()
-    if not res then return raise (err) end
+    if not res then
+        -- Errors in io.open have "filename: " prefix,
+        -- error in file:read don't, add it.
+        return raise (filename..": "..read_err)
+    end
     return res
 end
 
 --- write a string to a file
 -- @param filename The file path
 -- @param str The string
+-- @param is_bin open in binary mode
 -- @return true or nil
 -- @return error message
 -- @raise error if filename or str aren't strings
-function utils.writefile(filename,str)
+function utils.writefile(filename,str,is_bin)
+    local mode = is_bin and 'b' or ''
     utils.assert_string(1,filename)
     utils.assert_string(2,str)
-    local f,err = io.open(filename,'w')
+    local f,err = io.open(filename,'w'..mode)
     if not f then return raise(err) end
     f:write(str)
     f:close()
@@ -220,66 +226,36 @@ function utils.array_tostring (t,temp,tostr)
     return temp
 end
 
-if utils.lua51 then -- define Lua 5.2 style load()
-    local lua51_load = load
-    function utils.load(str,src,mode,env)
-        local chunk,err
-        if type(str) == 'string' then
-            chunk,err = loadstring(str,src)
-        else
-            chunk,err = lua51_load(str,src)
-        end
-        if chunk and env then setfenv(chunk,env) end
-        return chunk,err
-    end
-    utils.setfenv, utils.getfenv = setfenv, getfenv
-else
-    utils.load = load
-    -- setfenv/getfenv replacements for Lua 5.2
-    -- by Sergey Rozhenko
-    -- http://lua-users.org/lists/lua-l/2010-06/msg00313.html
-    -- Roberto Ierusalimschy notes that it is possible for getfenv to return nil
-    -- in the case of a function with no globals:
-    -- http://lua-users.org/lists/lua-l/2010-06/msg00315.html
-    function utils.setfenv(f, t)
-        f = (type(f) == 'function' and f or debug.getinfo(f + 1, 'f').func)
-        local name
-        local up = 0
-        repeat
-            up = up + 1
-            name = debug.getupvalue(f, up)
-        until name == '_ENV' or name == nil
-        if name then
-            debug.upvaluejoin(f, up, function() return name end, 1) -- use unique upvalue
-            debug.setupvalue(f, up, t)
-        end
-        if f ~= 0 then return f end
-    end
+local is_windows = utils.is_windows
 
-    function utils.getfenv(f)
-        local f = f or 0
-        f = (type(f) == 'function' and f or debug.getinfo(f + 1, 'f').func)
-        local name, val
-        local up = 0
-        repeat
-            up = up + 1
-            name, val = debug.getupvalue(f, up)
-        until name == '_ENV' or name == nil
-        return val
-    end
-end
+--- Quote an argument of a command.
+-- Quotes a single argument of a command to be passed
+-- to `os.execute`, `pl.utils.execute` or `pl.utils.executeex`.
+-- @string argument the argument.
+-- @return quoted argument.
+function utils.quote_arg(argument)
+    if is_windows then
+        if argument == "" or argument:find('[ \f\t\v]') then
+            -- Need to quote the argument.
+            -- Quotes need to be escaped with backslashes;
+            -- additionally, backslashes before a quote, escaped or not,
+            -- need to be doubled.
+            -- See documentation for CommandLineToArgvW Windows function.
+            argument = '"' .. argument:gsub([[(\*)"]], [[%1%1\"]]):gsub([[\+$]], "%0%0") .. '"'
+        end
 
---- execute a shell command.
--- This is a compatibility function that returns the same for Lua 5.1 and Lua 5.2
--- @param cmd a shell command
--- @return true if successful
--- @return actual return code
-function utils.execute (cmd)
-    local res1,res2,res2 = os.execute(cmd)
-    if lua51 then
-        return res1==0,res1
+        -- os.execute() uses system() C function, which on Windows passes command
+        -- to cmd.exe. Escape its special characters.
+        return (argument:gsub('["^<>!|&%%]', "^%0"))
     else
-        return res1,res2
+        if argument == "" or argument:find('[^a-zA-Z0-9_@%+=:,./-]') then
+            -- To quote arguments on posix-like systems use single quotes.
+            -- To represent an embedded single quote close quoted string ('),
+            -- add escaped quote (\'), open quoted string again (').
+            argument = "'" .. argument:gsub("'", [['\'']]) .. "'"
+        end
+
+        return argument
     end
 end
 
@@ -292,43 +268,23 @@ end
 -- @return stdout output (string)
 -- @return errout output (string)
 function utils.executeex(cmd, bin)
-  local mode
-  local outfile = os.tmpname()
-  local errfile = os.tmpname()
+    local mode
+    local outfile = os.tmpname()
+    local errfile = os.tmpname()
 
-  if is_windows then
-    outfile = os.getenv('TEMP')..outfile
-    errfile = os.getenv('TEMP')..errfile
-  end
-  cmd = cmd .. [[ >"]]..outfile..[[" 2>"]]..errfile..[["]]
-
-	local success, retcode = utils.execute(cmd)
-  local outcontent = utils.readfile(outfile, bin)
-  local errcontent = utils.readfile(errfile, bin)
-  os.remove(outfile)
-  os.remove(errfile)
-  return success, retcode, (outcontent or ""), (errcontent or "")
-end
-
-if not table.pack then
-    function table.pack (...)
-        return {n=select('#',...); ...}
+    if is_windows and not outfile:find(':') then
+        outfile = os.getenv('TEMP')..outfile
+        errfile = os.getenv('TEMP')..errfile
     end
-end
-if not package.searchpath then
-    local sep = package.config:sub(1,1)
-    function package.searchpath (mod,path)
-        mod = mod:gsub('%.',sep)
-        for m in path:gmatch('[^;]+') do
-            local nm = m:gsub('?',mod)
-            local f = io.open(nm,'r')
-            if f then f:close(); return nm end
-        end
-    end
-end
+    cmd = cmd .. " > " .. utils.quote_arg(outfile) .. " 2> " .. utils.quote_arg(errfile)
 
---if not table.pack then table.pack = _G.pack end
---if not rawget(_G,"pack") then _G.pack = table.pack end
+    local success, retcode = utils.execute(cmd)
+    local outcontent = utils.readfile(outfile, bin)
+    local errcontent = utils.readfile(errfile, bin)
+    os.remove(outfile)
+    os.remove(errfile)
+    return success, retcode, (outcontent or ""), (errcontent or "")
+end
 
 --- 'memoize' a function (cache returned value for next call).
 -- This is useful if you have a function which is relatively expensive,
@@ -337,73 +293,17 @@ end
 -- @param func a function of at least one argument
 -- @return a function with at least one argument, which is used as the key.
 function utils.memoize(func)
-    return setmetatable({}, {
-        __index = function(self, k, ...)
-            local v = func(k,...)
-            self[k] = v
-            return v
-        end,
-        __call = function(self, k) return self[k] end
-    })
-end
-
---- is the object either a function or a callable object?.
--- @param obj Object to check.
-function utils.is_callable (obj)
-    return type(obj) == 'function' or getmetatable(obj) and getmetatable(obj).__call
-end
-
---- is the object of the specified type?.
--- If the type is a string, then use type, otherwise compare with metatable
--- @param obj An object to check
--- @param tp String of what type it should be
-function utils.is_type (obj,tp)
-    if type(tp) == 'string' then return type(obj) == tp end
-    local mt = getmetatable(obj)
-    return tp == mt
-end
-
-local fileMT = getmetatable(io.stdout)
-
---- a string representation of a type.
--- For tables with metatables, we assume that the metatable has a `_name`
--- field. Knows about Lua file objects.
--- @param obj an object
--- @return a string like 'number', 'table' or 'List'
-function utils.type (obj)
-    local t = type(obj)
-    if t == 'table' or t == 'userdata' then
-        local mt = getmetatable(obj)
-        if mt == fileMT then
-            return 'file'
-        else
-            return mt._name or "unknown "..t
+    local cache = {}
+    return function(k)
+        local res = cache[k]
+        if res == nil then
+            res = func(k)
+            cache[k] = res
         end
-    else
-        return t
+        return res
     end
 end
 
---- is this number an integer?
--- @param x a number
--- @raise error if x is not a number
-function utils.is_integer (x)
-    return math.ceil(x)==x
-end
-
---- Check if the object is "empty".
--- An object is considered empty if it is nil, a table with out any items (key,
--- value pairs or indexes), or a string with no content ("").
--- @param o The object to check if it is empty.
--- @param ignore_spaces If the object is a string and this is true the string is
--- considered empty is it only contains spaces.
--- @return true if the object is empty, otherwise false.
-function utils.is_empty(o, ignore_spaces)
-	if o == nil or (type(o) == "table" and not next(o)) or (type(o) == "string" and (o == "" or (ignore_spaces and o:match("^%s+$")))) then
-		return true
-	end
-	return false
-end
 
 utils.stdmt = {
     List = {_name='List'}, Map = {_name='Map'},
@@ -415,8 +315,8 @@ local _function_factories = {}
 --- associate a function factory with a type.
 -- A function factory takes an object of the given type and
 -- returns a function for evaluating it
--- @param mt metatable
--- @param fun a callable that returns a function
+-- @tab mt metatable
+-- @func fun a callable that returns a function
 function utils.add_function_factory (mt,fun)
     _function_factories[mt] = fun
 end
@@ -432,7 +332,7 @@ local function _string_lambda(f)
             if not args then return raise 'bad string lambda' end
         end
         local fstr = 'return function('..args..') return '..body..' end'
-        local fn,err = loadstring(fstr)
+        local fn,err = utils.load(fstr)
         if not fn then return raise(err) end
         fn = fn()
         return fn
@@ -445,7 +345,7 @@ end
 -- @param lf function as a string
 -- @return a function
 -- @usage string_lambda '|x|x+1' (2) == 3
--- @usage string_lambda '_+1 (2) == 3
+-- @usage string_lambda '_+1' (2) == 3
 -- @function utils.string_lambda
 utils.string_lambda = utils.memoize(_string_lambda)
 
@@ -461,7 +361,6 @@ local ops
 -- @param msg optional error message
 -- @return a callable
 -- @raise if idx is not a number or if f is not callable
--- @see utils.is_callable
 function utils.function_arg (idx,f,msg)
     utils.assert_arg(1,idx,'number')
     local tp = type(f)
@@ -498,7 +397,7 @@ end
 -- @param p a value
 -- @return a function such that f(x) is fn(p,x)
 -- @raise same as @{function_arg}
--- @see pl.func.curry
+-- @see func.bind1
 function utils.bind1 (fn,p)
     fn = utils.function_arg(1,fn)
     return function(...) return fn(p,...) end
@@ -519,7 +418,7 @@ end
 -- @param n argument index
 -- @param val the value
 -- @param tp the type
--- @param verify an optional verfication function
+-- @param verify an optional verification function
 -- @param msg an optional custom message
 -- @param lev optional stack position for trace, default 2
 -- @raise if the argument n is not the correct type
@@ -572,6 +471,16 @@ function utils.raise (err)
     end
 end
 
+--- is the object of the specified type?.
+-- If the type is a string, then use type, otherwise compare with metatable
+-- @param obj An object to check
+-- @param tp String of what type it should be
+function utils.is_type (obj,tp)
+    if type(tp) == 'string' then return type(obj) == tp end
+    local mt = getmetatable(obj)
+    return tp == mt
+end
+
 raise = utils.raise
 
 --- load a code string or bytecode chunk.
@@ -583,22 +492,25 @@ raise = utils.raise
 -- @return error message (chunk is nil)
 -- @function utils.load
 
+---------------
+-- Get environment of a function.
+-- With Lua 5.2, may return nil for a function with no global references!
+-- Based on code by [Sergey Rozhenko](http://lua-users.org/lists/lua-l/2010-06/msg00313.html)
+-- @param f a function or a call stack reference
+-- @function utils.getfenv
 
---- Lua 5.2 Compatible Functions
--- @section lua52
+---------------
+-- Set environment of a function
+-- @param f a function or a call stack reference
+-- @param env a table that becomes the new environment of `f`
+-- @function utils.setfenv
 
---- pack an argument list into a table.
--- @param ... any arguments
--- @return a table with field n set to the length
--- @return the length
--- @function table.pack
-
-------
--- return the full path where a Lua module name would be matched.
--- @param mod module name, possibly dotted
--- @param path a path in the same form as package.path or package.cpath
--- @see path.package_path
--- @function package.searchpath
+--- execute a shell command.
+-- This is a compatibility function that returns the same for Lua 5.1 and Lua 5.2
+-- @param cmd a shell command
+-- @return true if successful
+-- @return actual return code
+-- @function utils.execute
 
 return utils
 
